/*
 * Created on 14-abr-2005
 *
 * gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 * DwgMemoryDriver 0.2. Driver del formato DWG para gvSIG
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
package es.iver.gvsig.drivers.dwg;

import java.awt.Color;
import java.awt.Font;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Vector;

import net.jmorell.javacad.util.AcadColor;
import net.jmorell.javacad.util.GisModelCurveCalculator;
import net.jmorell.jdwglib.dwg.DwgFile;
import net.jmorell.jdwglib.dwg.DwgObject;
import net.jmorell.jdwglib.dwg.objects.DwgArc;
import net.jmorell.jdwglib.dwg.objects.DwgAttdef;
import net.jmorell.jdwglib.dwg.objects.DwgAttrib;
import net.jmorell.jdwglib.dwg.objects.DwgBlock;
import net.jmorell.jdwglib.dwg.objects.DwgBlockControl;
import net.jmorell.jdwglib.dwg.objects.DwgBlockHeader;
import net.jmorell.jdwglib.dwg.objects.DwgCircle;
import net.jmorell.jdwglib.dwg.objects.DwgEndblk;
import net.jmorell.jdwglib.dwg.objects.DwgInsert;
import net.jmorell.jdwglib.dwg.objects.DwgLine;
import net.jmorell.jdwglib.dwg.objects.DwgLwPolyline;
import net.jmorell.jdwglib.dwg.objects.DwgMText;
import net.jmorell.jdwglib.dwg.objects.DwgPoint;
import net.jmorell.jdwglib.dwg.objects.DwgPolyline2D;
import net.jmorell.jdwglib.dwg.objects.DwgPolyline3D;
import net.jmorell.jdwglib.dwg.objects.DwgSolid;
import net.jmorell.jdwglib.dwg.objects.DwgText;
import net.jmorell.jdwglib.dwg.objects.DwgEllipse;

import com.hardcode.gdbms.engine.data.driver.DriverException;
import com.hardcode.gdbms.engine.data.driver.ObjectDriver;
import com.hardcode.gdbms.engine.data.edition.DataWare;
import com.hardcode.gdbms.engine.values.IntValue;
import com.hardcode.gdbms.engine.values.Value;
import com.hardcode.gdbms.engine.values.ValueFactory;
import com.iver.cit.gvsig.fmap.core.FPoint2D;
import com.iver.cit.gvsig.fmap.core.FPoint3D;
import com.iver.cit.gvsig.fmap.core.FPolygon2D;
import com.iver.cit.gvsig.fmap.core.FPolyline2D;
import com.iver.cit.gvsig.fmap.core.FPolyline3D;
import com.iver.cit.gvsig.fmap.core.FShape;
import com.iver.cit.gvsig.fmap.core.GeneralPathX;
import com.iver.cit.gvsig.fmap.core.v02.FConstant;
import com.iver.cit.gvsig.fmap.core.v02.FSymbol;
import com.iver.cit.gvsig.fmap.drivers.DriverAttributes;
import com.iver.cit.gvsig.fmap.drivers.MemoryDriver;
import com.iver.cit.gvsig.fmap.drivers.WithDefaultLegend;
import com.iver.cit.gvsig.fmap.rendering.Legend;
import com.iver.cit.gvsig.fmap.rendering.LegendFactory;
import com.iver.cit.gvsig.fmap.rendering.VectorialUniqueValueLegend;

/**
 * Driver that allows gvSig to read files in DWG format
 * Using this driver, a gvSig user can manipulate part of the information contained in
 * a DWG file
 * This driver load the Dwg file in memory
 * This driver uses jdwglib
 * 
 * @author jmorell
 */
public class DwgMemoryDriver extends MemoryDriver implements WithDefaultLegend {
	private final int ID_FIELD_ID=0;
	private final int ID_FIELD_FSHAPE=1;
	private final int ID_FIELD_ENTITY=2;
	private final int ID_FIELD_LAYER=3;
	private final int ID_FIELD_COLOR=4;
	private final int ID_FIELD_ELEVATION=5;
	private final int ID_FIELD_THICKNESS=6;
	private final int ID_FIELD_TEXT=7;
	private final int ID_FIELD_HEIGHTTEXT=8;
	private final int ID_FIELD_ROTATIONTEXT=9;

	VectorialUniqueValueLegend defaultLegend;
	private String path;
	private File m_Fich;
	
	private DwgFile dwg;
	private Vector dwgObjects;
	private DriverAttributes attr = new DriverAttributes();

	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.drivers.MemoryDriver#open(java.io.File)
	 */
	public void open(File f) throws IOException {
    	m_Fich = f;
	}
	
	/**
	 * This method load the DWG file in memory.
	 * First, it will be necessary to create a DwgFile object with the DWG file path
	 * as the argument
	 * Second, the method read of DwgFile allows to read the objects inside the DWG
	 * file
	 * Third, it will be necessary to process some DWG objects like Layers or Polylines
	 * Fourth, we can read the objects Vector, and convert this objects to the FMap
	 * object model
	 * And finally, the testDwg3D method test if this DWG has elevation informacion
	 */
	public void initialize() throws IOException {
		float heightText = 10;
		
		attr.setLoadedInMemory(true);
		
		dwg = new DwgFile(m_Fich.getAbsolutePath());
		long t1 = System.currentTimeMillis();
		dwg.read();
		long t2 = System.currentTimeMillis();
		System.out.println("Tiempo empleado por la librera en leer el fichero dwg = " + (t2-t1));
		t1 = System.currentTimeMillis();
		dwg.initializeLayerTable();
        dwg.calculateGisModelDwgPolylines();
        dwg.applyExtrusions();
		t2 = System.currentTimeMillis();
		System.out.println("Tiempo empleado por la librera en tratar el fichero dwg = " + (t2-t1));
		dwgObjects = dwg.getDwgObjects();
		dwg.testDwg3D();
		
		// TODO: Atributos extra de los objectos
		//int nAtt = featureMaker.getAttributes().size();
		
		// Campos de las MemoryLayer:
    	//Value[] auxRow = new Value[9+nAtt];
    	Value[] auxRow = new Value[10];
		ArrayList arrayFields = new ArrayList();
		arrayFields.add("ID");
		arrayFields.add("FShape");
		arrayFields.add("Entity");
		arrayFields.add("Layer");
		arrayFields.add("Color");
		arrayFields.add("Elevation");
		arrayFields.add("Thickness");
		arrayFields.add("Text");
		arrayFields.add("HeightText");
		arrayFields.add("RotationText");
		/*for (int i=0;i<nAtt;i++) {
			String att[] = new String[2];
			att = (String[])featureMaker.getAttributes().get(i);
			arrayFields.add(att[0]);
		}*/
		
    	getTableModel().setColumnIdentifiers(arrayFields.toArray());
    	
		t1 = System.currentTimeMillis();
		
    	boolean addingToBlock = false;
    	
		for (int i=0;i<dwgObjects.size();i++) {
			
            auxRow[ID_FIELD_HEIGHTTEXT] = ValueFactory.createValue(0.0);
            auxRow[ID_FIELD_ROTATIONTEXT] = ValueFactory.createValue(0.0);
            auxRow[ID_FIELD_TEXT] = ValueFactory.createNullValue();
			
			DwgObject entity = (DwgObject)dwgObjects.get(i);
			if (entity instanceof DwgArc && !addingToBlock) {
				//System.out.println("Encuentra un Arc ...");
				double[] c = ((DwgArc)entity).getCenter();
				Point2D center = new Point2D.Double(c[0], c[1]);
				double radius = ((DwgArc)entity).getRadius();
				double initAngle = Math.toDegrees(((DwgArc)entity).getInitAngle());
				double endAngle = Math.toDegrees(((DwgArc)entity).getEndAngle());
				Point2D[] arc = GisModelCurveCalculator.calculateGisModelArc(center, radius, initAngle, endAngle);
				FPolyline2D arcc;
				if (dwg.isDwg3DFile()) {
				    double[][] arc3D = new double[arc.length][3];
				    for (int j=0;j<arc.length;j++) {
				        arc3D[j][0] = arc[j].getX();
				        arc3D[j][1] = arc[j].getY();
				        arc3D[j][2] = c[2];
				    }
				    arcc = points3DToFPolyline3D(arc3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else {
				    arcc = points2DToFPolyline2D(arc);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Arc"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				// TODO: if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(c[2]);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(arcc, auxRow);
			} else if (entity instanceof DwgEllipse && !addingToBlock) {
				//System.out.println("Encuentra una Ellipse ...");
				double[] c = ((DwgEllipse)entity).getCenter();
				Point2D center = new Point2D.Double(c[0], c[1]);
				double[] majorAxisVector = ((DwgEllipse)entity).getMajorAxisVector();
				Point2D mav = new Point2D.Double(majorAxisVector[0], majorAxisVector[1]);
				double axisRatio = ((DwgEllipse)entity).getAxisRatio();
				double initAngle = Math.toDegrees(((DwgEllipse)entity).getInitAngle());
				double endAngle = Math.toDegrees(((DwgEllipse)entity).getEndAngle());
				Point2D[] arc = GisModelCurveCalculator.calculateGisModelEllipse(center, mav, axisRatio, initAngle, endAngle);
				FPolyline2D arcc;
				if (dwg.isDwg3DFile()) {
				    double[][] arc3D = new double[arc.length][3];
				    for (int j=0;j<arc.length;j++) {
				        arc3D[j][0] = arc[j].getX();
				        arc3D[j][1] = arc[j].getY();
				        arc3D[j][2] = c[2];
				    }
				    arcc = points3DToFPolyline3D(arc3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else {
				    arcc = points2DToFPolyline2D(arc);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Ellipse"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				// TODO: if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(c[2]);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(arcc, auxRow);
			} else if (entity instanceof DwgCircle && !addingToBlock) {
				//System.out.println("Encuentra un Circle ...");
				double[] c = ((DwgCircle)entity).getCenter();
				Point2D center = new Point2D.Double(c[0], c[1]);
				double radius = ((DwgCircle)entity).getRadius();
				Point2D[] arc = GisModelCurveCalculator.calculateGisModelCircle(center, radius);
				FPolyline2D arcc;
				if (dwg.isDwg3DFile()) {
				    double[][] arc3D = new double[arc.length][3];
				    for (int j=0;j<arc.length;j++) {
				        arc3D[j][0] = arc[j].getX();
				        arc3D[j][1] = arc[j].getY();
				        arc3D[j][2] = c[2];
				    }
				    arcc = points3DToFPolyline3D(arc3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else {
				    arcc = points2DToFPolyline2D(arc);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Circle"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(c[2]);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(arcc, auxRow);
			} else if (entity instanceof DwgPolyline2D && !addingToBlock) {
				//System.out.println("Encuentra una Polyline2D ...");
				Point2D[] points = ((DwgPolyline2D)entity).getPts();
				double elev = ((DwgPolyline2D)entity).getElevation();
				FPolyline2D pline;
				if (points!=null) {
					if (dwg.isDwg3DFile()) {
					    double[][] pline3D = new double[points.length][3];
					    for (int j=0;j<points.length;j++) {
					        pline3D[j][0] = points[j].getX();
					        pline3D[j][1] = points[j].getY();
					        pline3D[j][2] = elev;
					    }
					    pline = points3DToFPolyline3D(pline3D);
		                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
					} else {
					    pline = points2DToFPolyline2D(points);
		                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
					}
					auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
	                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Polyline2D"));
					String layerName = dwg.getLayerName(entity);
	                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
					int color = entity.getColor();
					int colorByLayer = dwg.getColorByLayer(entity);
					//if (color==0) color ByBlock
					if (color==256) color = colorByLayer;
	                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
	            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(elev);
	            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
	            	// Attributes
	            	/*for (int j=0;j<nAtt;j++) {
	    				String[] attributes = new String[2];
	    				attributes = (String[])featureMaker.getAttributes().get(j);
	                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
	                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
	                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
	                	}
	            	}*/
	            	addShape(pline, auxRow);
				}
			} else if (entity instanceof DwgPolyline3D && !addingToBlock) {
				//System.out.println("Encuentra una Polyline3D ...");
				double[][] points3D = ((DwgPolyline3D)entity).getPts();
				FPolyline2D pline; 
				if (points3D!=null) {
					if (dwg.isDwg3DFile()) {
					    double[][] pline3D = new double[points3D.length][3];
					    pline = points3DToFPolyline3D(points3D);
		                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
					} else {
						Point2D[] points2D = new Point2D[points3D.length];
					    for (int j=0;j<points3D.length;j++) {
					        points2D[j] = new Point2D.Double(points3D[j][0], points3D[j][1]);
					    }
					    pline = points2DToFPolyline2D(points2D);
		                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
					}
					auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
	                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Polyline3D"));
					String layerName = dwg.getLayerName(entity);
	                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
					int color = entity.getColor();
					int colorByLayer = dwg.getColorByLayer(entity);
					//if (color==0) color ByBlock
					if (color==256) color = colorByLayer;
	                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
	                boolean constantElevation = true;
	                for (int j=0;j<points3D.length;j++) {
	                    if (points3D[j][2]!=points3D[0][2]) {
	                        constantElevation = false;
	                        break;
	                    }
	                }
	            	if (constantElevation) auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(points3D[0][2]);
	            	else auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
	            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
	            	// Attributes
	            	/*for (int j=0;j<nAtt;j++) {
	    				String[] attributes = new String[2];
	    				attributes = (String[])featureMaker.getAttributes().get(j);
	                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
	                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
	                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
	                	}
	            	}*/
	            	addShape(pline, auxRow);
				}
			} else if (entity instanceof DwgLwPolyline && !addingToBlock) {
				//System.out.println("Encuentra una Lwpolyline ...");
				Point2D[] pts = ((DwgLwPolyline)entity).getVertices();
				double elev = ((DwgLwPolyline)entity).getElevation();
				if (pts!=null) {
					FPolyline2D lwpline;
					if (dwg.isDwg3DFile()) {
					    double[][] pline3D = new double[pts.length][3];
					    for (int j=0;j<pts.length;j++) {
					        pline3D[j][0] = pts[j].getX();
					        pline3D[j][1] = pts[j].getY();
					        pline3D[j][2] = elev;
					    }
					    lwpline = points3DToFPolyline3D(pline3D);
		                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
					} else {
					    lwpline = points2DToFPolyline2D(pts);
		                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
					}
	                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
	                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("LwPolyline"));
					String layerName = dwg.getLayerName(entity);
	                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
					int color = entity.getColor();
					int colorByLayer = dwg.getColorByLayer(entity);
					//if (color==0) color ByBlock
					if (color==256) color = colorByLayer;
	                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
	            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(elev);
	            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
	            	// Attributes
	            	/*for (int j=0;j<nAtt;j++) {
	    				String[] attributes = new String[2];
	    				attributes = (String[])featureMaker.getAttributes().get(j);
	                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
	                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
	                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
	                	}
	            	}*/
					addShape(lwpline, auxRow);
				}
			} else if (entity instanceof DwgSolid && !addingToBlock) {
				//System.out.println("Encuentra un Solid ...");
				double[] p1 = ((DwgSolid)entity).getCorner1();
				double[] p2 = ((DwgSolid)entity).getCorner2();
				double[] p3 = ((DwgSolid)entity).getCorner3();
				double[] p4 = ((DwgSolid)entity).getCorner4();
				double elev = ((DwgSolid)entity).getElevation();
				Point2D[] pts = new Point2D[]{new Point2D.Double(p1[0], p1[1]), new Point2D.Double(p2[0], p2[1]), new Point2D.Double(p4[0], p4[1]), new Point2D.Double(p3[0], p3[1])};
				FPolyline2D solid;
				if (dwg.isDwg3DFile()) {
				    double[][] pline3D = new double[pts.length][3];
				    for (int j=0;j<pts.length;j++) {
				        pline3D[j][0] = pts[j].getX();
				        pline3D[j][1] = pts[j].getY();
				        pline3D[j][2] = elev;
				    }
				    solid = points3DToFPolyline3D(pline3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else {
				    solid = points2DToFPolyline2D(pts);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Solid"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(elev);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(solid, auxRow);
			} else if (entity instanceof DwgLine && !addingToBlock) {
				//System.out.println("Encuentra una Line ...");
				double[] p1 = ((DwgLine)entity).getP1();
				Point2D point1 = new Point2D.Double(p1[0], p1[1]);
				double[] p2 = ((DwgLine)entity).getP2();
				Point2D point2 = new Point2D.Double(p2[0], p2[1]);
				Point2D[] lin = new Point2D[]{point1, point2};
				FPolyline2D line;
				if (dwg.isDwg3DFile() && !((DwgLine)entity).isZflag()) {
				    double[][] lin3D = new double[2][3];
			        lin3D[0][0] = lin[0].getX();
			        lin3D[0][1] = lin[0].getY();
			        lin3D[0][2] = p1[2];
			        lin3D[1][0] = lin[1].getX();
			        lin3D[1][1] = lin[1].getY();
			        lin3D[1][2] = p2[2];
				    line = points3DToFPolyline3D(lin3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else if (dwg.isDwg3DFile() && ((DwgLine)entity).isZflag()) {
				    double[][] lin3D = new double[2][3];
			        lin3D[0][0] = lin[0].getX();
			        lin3D[0][1] = lin[0].getY();
			        lin3D[0][2] = 0.0;
			        lin3D[1][0] = lin[1].getX();
			        lin3D[1][1] = lin[1].getY();
			        lin3D[1][2] = 0.0;
				    line = points3DToFPolyline3D(lin3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else {
				    line = points2DToFPolyline2D(lin);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Line"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	if (!((DwgLine)entity).isZflag()) {
                    if (p1[2]==p2[2]) auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(p1[2]);
                	else auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
            	} else {
                	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
            	}
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(line, auxRow);
			} else if (entity instanceof DwgPoint && !addingToBlock) {
				//System.out.println("Encuentra un Point ...");
				double[] p = ((DwgPoint)entity).getPoint();
				FPoint2D point;
				if (dwg.isDwg3DFile()) {
				    point = new FPoint3D(p[0], p[1], p[2]);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint3D"));
				} else {
				    point = new FPoint2D(p[0], p[1]);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Point"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(p[2]);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(point, auxRow);
			} else if (entity instanceof DwgInsert && !addingToBlock) {
				//System.out.println("Encuentra un Insert ...");
				double[] p = ((DwgInsert)entity).getInsertionPoint();
				Point2D point = new Point2D.Double(p[0], p[1]);
				FPoint2D fPoint;
				double[] scale = ((DwgInsert)entity).getScale();
				double rot = ((DwgInsert)entity).getRotation();
				int blockHandle = ((DwgInsert)entity).getBlockHeaderHandle();
				manageInsert(dwgObjects, point, scale, rot, blockHandle, i, auxRow);
				if (dwg.isDwg3DFile()) {
				    fPoint = new FPoint3D(p[0], p[1], p[2]);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint3D"));
				} else {
				    fPoint = new FPoint2D(p[0], p[1]);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint2D"));
				}
				auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Insert"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(p[2]);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(fPoint, auxRow);
			} else if (entity instanceof DwgMText && !addingToBlock) {
				//System.out.println("Encuentra un MText ...");
				double[] p = ((DwgMText)entity).getInsertionPoint();
				FPoint2D fPoint;
				if (dwg.isDwg3DFile()) {
				    fPoint = new FPoint3D(p[0], p[1], p[2]);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint3D"));
				} else {
				    fPoint = new FPoint2D(p[0], p[1]);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("MText"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(p[2]);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(fPoint, auxRow);
			} else if (entity instanceof DwgAttrib && !addingToBlock) {
				//System.out.println("Encuentra un Attrib ...");
			} else if (entity instanceof DwgText && !addingToBlock) {
				//System.out.println("Encuentra un Text ...");
				Point2D p = ((DwgText)entity).getInsertionPoint();
				int dflag = ((DwgText)entity).getDataFlag();
				double elev = 0.0;
				if ((dflag & 0x1)==0) elev = ((DwgText)entity).getElevation();
				FPoint2D point;
				if (dwg.isDwg3DFile()) {
				    point = new FPoint3D(p.getX(), p.getY(), elev);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint3D"));
				} else {
				    point = new FPoint2D(p.getX(), p.getY());
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPoint2D"));
				}
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Text"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
                String text = ((DwgText)entity).getText();
                auxRow[ID_FIELD_TEXT] = ValueFactory.createValue(new String(text));
                heightText = (float)((DwgText)entity).getHeight();
                auxRow[ID_FIELD_HEIGHTTEXT] = ValueFactory.createValue(heightText);
                if ((dflag & 0x8) == 0) {
	            	double textRot = ((DwgText)entity).getRotationAngle();
                	auxRow[ID_FIELD_ROTATIONTEXT] = ValueFactory.createValue(Math.toDegrees(textRot));
				} else {
	            	auxRow[ID_FIELD_ROTATIONTEXT] = ValueFactory.createValue(0.0);
				}
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(elev);
            	double thickness = ((DwgText)entity).getThickness();
				auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(thickness);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				addShape(point, auxRow);
			} else if (entity instanceof DwgBlock) {
				//System.out.println("Encuentra un Block ...");
				addingToBlock = true;
				String blockName = ((DwgBlock)entity).getName();
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Block"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				//addShape(point, auxRow);
			} else if (entity instanceof DwgEndblk) {
				//System.out.println("Encuentra un Endblk ...");
				addingToBlock = false;
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("EndBlk"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				//addShape(point, auxRow);
			} else if (entity instanceof DwgBlockControl) {
				//System.out.println("Encuentra un Block Control Object ...");
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Block Control Object"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				//addShape(point, auxRow);
			} else if (entity instanceof DwgBlockHeader) {
				//System.out.println("Encuentra un Block Header ...");
				addingToBlock = true;
                auxRow[ID_FIELD_ID] = ValueFactory.createValue(i);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Block Header"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
				//addShape(point, auxRow);
			} else if (entity instanceof DwgAttdef && !addingToBlock) {
				//System.out.println("Encuentra un Attdef ...");
			} else {
				//System.out.println("Detectado dwgObject pendiente de implementar");
			}
		}
		t2 = System.currentTimeMillis();
		System.out.println("Tiempo empleado por el driver en importar el contenido del dwg = " + (t2-t1));
		
		defaultLegend = LegendFactory.createVectorialUniqueValueLegend(getShapeType());
		defaultLegend.setFieldName("Color");
		defaultLegend.setLabelField("Text");
		defaultLegend.setDefaultSymbol(new FSymbol(getShapeType()));
		defaultLegend.getDefaultSymbol().setShapeVisible(false);
		defaultLegend.getDefaultSymbol().setFontSizeInPixels(false);
		defaultLegend.getDefaultSymbol().setFont(new Font("SansSerif", Font.PLAIN, 9));
		defaultLegend.getDefaultSymbol().setFontColor(Color.BLACK);
		defaultLegend.getDefaultSymbol().setFontSize(heightText);
		defaultLegend.getDefaultSymbol().setStyle(FConstant.SYMBOL_STYLE_DGNSPECIAL);
		defaultLegend.getDefaultSymbol().setSize(3);
		defaultLegend.getDefaultSymbol().setSizeInPixels(true);
		
		defaultLegend.setLabelHeightField("HeightText");
		defaultLegend.setLabelRotationField("RotationText");
		
        ObjectDriver rs = this;
		IntValue clave;
		FSymbol theSymbol = null;
		
		try {
		    // Provisional hasta que cambiemos los smbolos.
		   /*  BufferedImage bi= new BufferedImage(5, 5, BufferedImage.TYPE_INT_ARGB);
		   Graphics2D big = bi.createGraphics();
		   Color color=new Color(0,0,0,0);
		   big.setBackground(color);
		   big.clearRect(0, 0, 5, 5);
		   Paint fillProv = null; 
		   Rectangle2D rProv = new Rectangle();
		   rProv.setFrame(0, 0,5,5);
		   fillProv = new TexturePaint(bi,rProv); */		    
		    
			for (long j = 0; j < rs.getRowCount(); j++)
			{
				clave = (IntValue)rs.getFieldValue(j,ID_FIELD_COLOR);
				if (defaultLegend.getSymbolByValue(clave) == null)
				{
					theSymbol = new FSymbol(getShapeType());
					theSymbol.setDescription(clave.toString());
					theSymbol.setColor(AcadColor.getColor(clave.getValue()));
					// theSymbol.setFill(fillProv);
					// 050202, jmorell: Asigna los colores de Autocad a los bordes
					//					de los polgonos.
					theSymbol.setOutlineColor(AcadColor.getColor(clave.getValue()));
					
					theSymbol.setStyle(FConstant.SYMBOL_STYLE_DGNSPECIAL);
					theSymbol.setSize(3);
					theSymbol.setSizeInPixels(true);
					
					defaultLegend.addSymbol(clave, theSymbol);
				}
			} // for
		} catch (DriverException e) {
			e.printStackTrace();
			throw new IOException("Error al poner la leyenda por defecto en el Dwg");
		}
	}
	
	/**
	 * Method that changes a Point2D array to a FPolyline2D. Is useful to convert a
	 * polyline given by it points to a FPolyline2D, a polyline in the FMap model
	 * object
	 * 
	 * @param pts Array of Point2D that defines the polyline that will be converted in
	 * 		  a FPolyline2D
	 * @return FPolyline2D This FPolyline2D is build using the array of Point2D that
	 * 		  is the argument of the method
	 */
	private FPolyline2D points2DToFPolyline2D(Point2D[] pts) {
		GeneralPathX genPathX = new GeneralPathX();
		genPathX.moveTo(pts[0].getX(), pts[0].getY());
		for (int i=1; i<pts.length; i++) {
			genPathX.lineTo(pts[i].getX(), pts[i].getY());
		}
		return new FPolyline2D(genPathX);
	}
	
	/**
	 * Method that changes a Point3D array to a FPolyline3D. Is useful to convert a
	 * polyline given by it points to a FPolyline3D, a polyline 3D in the FMap model
	 * object
	 * 
	 * @param pts Array of Point3D that defines the polyline 3D that will be converted in
	 * 		  a FPolyline3D
	 * @return FPolyline3D This FPolyline3D is build using the array of Point3D that
	 * 		  is the argument of the method
	 */
	private FPolyline3D points3DToFPolyline3D(double[][] pts) {
		GeneralPathX genPathX = new GeneralPathX();
		genPathX.moveTo(pts[0][0], pts[0][1]);
		for (int i=1; i<pts.length; i++) {
			genPathX.lineTo(pts[i][0], pts[i][1]);
		}
		double[] elevations = new double[pts.length];
		for (int i=0;i<pts.length;i++) {
		    elevations[i] = pts[i][2];
		}
		return new FPolyline3D(genPathX, elevations);
	}
	
	/**
	 * Method that changes a Point2D array to a FPolygon2D. Is useful to convert a
	 * polygon given by it points to a FPolygon2D, a polygon in the FMap model
	 * object
	 * 
	 * @param pts Array of Point2D that defines the polygon that will be converted in
	 * 		  a FPolygon2D
	 * @return FPolygon2D This FPolygon2D is build using the array of Point2D that
	 * 		  is the argument of the method
	 */
	private FPolygon2D points2DToFPolygon2D(Point2D[] pts) {
		GeneralPathX genPathX = new GeneralPathX();
		genPathX.moveTo(pts[0].getX(), pts[0].getY());
		for (int i=1; i<pts.length; i++) {
			genPathX.lineTo(pts[i].getX(), pts[i].getY());
		}
		genPathX.closePath();
		return new FPolygon2D(genPathX);
	}
	
	private void manageInsert(Vector dwgObjects, Point2D insPoint, double[] scale, double rot, int bHandle, int id, Value[] auxRow) {
	    for (int i=0;i<dwgObjects.size();i++) {
			DwgObject obj = (DwgObject)dwgObjects.get(i);
			if (obj instanceof DwgBlockHeader) {
				int objHandle = ((DwgBlockHeader)obj).getHandle();
				if (objHandle==bHandle) {
					double[] bPoint = ((DwgBlockHeader)obj).getBasePoint();
					String bname = ((DwgBlockHeader)obj).getName();
					if (!bname.startsWith("*")) {
						int firstObjectHandle = ((DwgBlockHeader)obj).getFirstEntityHandle();
						int lastObjectHandle = ((DwgBlockHeader)obj).getLastEntityHandle();
						DwgBlock block = null;
						for (int j=0;j<dwgObjects.size();j++) {
							DwgObject ent = (DwgObject)dwgObjects.get(j);
							if (ent instanceof DwgBlock) {
								String name = ((DwgBlock)ent).getName();
								if (bname.equals(name)) {
									block = (DwgBlock)ent;
									break;
								}
							}
						}
						for (int j=0;j<dwgObjects.size();j++) {
							DwgObject fObj = (DwgObject)dwgObjects.get(j);
							if (fObj!=null) {
								int fObjHandle = fObj.getHandle();
								if (fObjHandle==firstObjectHandle) {
									int k=0;
									while (true) {
										DwgObject iObj = (DwgObject)dwgObjects.get(j+k);
										int iObjHandle = iObj.getHandle();
										manageBlockEntity(iObj, bPoint, insPoint, scale, rot, block, id, auxRow);
										k++;
										if (iObjHandle==lastObjectHandle) break;
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
	private void manageBlockEntity(DwgObject entity, double[] bPoint, Point2D insPoint, double[] scale, double rot, DwgObject block, int id, Value[] auxRow) {
		if (entity instanceof DwgArc) {
			//System.out.println("Encuentra un arco dentro de un bloque ...");
			double[] c = ((DwgArc)entity).getCenter();
			Point2D center = new Point2D.Double(c[0], c[1]);
			double radius = ((DwgArc)entity).getRadius();
			double initAngle = Math.toDegrees(((DwgArc)entity).getInitAngle());
			double endAngle = Math.toDegrees(((DwgArc)entity).getEndAngle());
			Point2D[] arc = GisModelCurveCalculator.calculateGisModelArc(center, radius, initAngle, endAngle);
			for (int i=0;i<arc.length;i++) {
				Point2D pointAux = new Point2D.Double(arc[i].getX() - bPoint[0], arc[i].getY() - bPoint[1]);
				double laX = insPoint.getX() + ((pointAux.getX()*scale[0])*Math.cos(rot) + (pointAux.getY()*scale[1])*(-1)*Math.sin(rot));
				double laY = insPoint.getY() + ((pointAux.getX()*scale[0])*Math.sin(rot) + (pointAux.getY()*scale[1])*Math.cos(rot));
				arc[i] = new Point2D.Double(laX, laY);
			}
			FPolyline2D arcc;
			if (dwg.isDwg3DFile()) {
			    double[][] arc3D = new double[arc.length][3];
			    for (int j=0;j<arc.length;j++) {
			        arc3D[j][0] = arc[j].getX();
			        arc3D[j][1] = arc[j].getY();
			        arc3D[j][2] = c[2];
			    }
			    arcc = points3DToFPolyline3D(arc3D);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
			} else {
			    arcc = points2DToFPolyline2D(arc);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
			}
            auxRow[ID_FIELD_ID] = ValueFactory.createValue(id);
            auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Arc"));
			String layerName = dwg.getLayerName(entity);
            auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
			int color = entity.getColor();
			int colorByLayer = dwg.getColorByLayer(entity);
			// TODO: if (color==0) color ByBlock
			if (color==256) color = colorByLayer;
            auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
        	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(c[2]);
        	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
        	// Attributes
        	//for (int j=0;j<nAtt;j++) {
			//	String[] attributes = new String[2];
			//	attributes = (String[])featureMaker.getAttributes().get(j);
            //	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
            //	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
            //		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
            //	}
        	//}
			addShape(arcc, auxRow);
		} else if (entity instanceof DwgCircle) {
			//System.out.println("Encuentra un crculo dentro de un bloque ...");
			double[] c = ((DwgCircle)entity).getCenter();
			Point2D center = new Point2D.Double(c[0], c[1]);
			double radius = ((DwgCircle)entity).getRadius();
			Point2D pointAux = new Point2D.Double(c[0] - bPoint[0], c[1] - bPoint[1]);
			double laX = insPoint.getX() + ((pointAux.getX()*scale[0])*Math.cos(Math.toRadians(rot)) + (pointAux.getY()*scale[1])*(-1)*Math.sin(Math.toRadians(rot)));
			double laY = insPoint.getY() + ((pointAux.getX()*scale[0])*Math.sin(Math.toRadians(rot)) + (pointAux.getY()*scale[1])*Math.cos(Math.toRadians(rot)));
			Point2D cen = new Point2D.Double(laX, laY);
			radius = radius * scale[0];
			Point2D[] arc = GisModelCurveCalculator.calculateGisModelCircle(new Point2D.Double(cen.getX(), cen.getY()), radius);
			FPolyline2D arcc;
			if (dwg.isDwg3DFile()) {
			    double[][] arc3D = new double[arc.length][3];
			    for (int j=0;j<arc.length;j++) {
			        arc3D[j][0] = arc[j].getX();
			        arc3D[j][1] = arc[j].getY();
			        arc3D[j][2] = c[2];
			    }
			    arcc = points3DToFPolyline3D(arc3D);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
			} else {
			    arcc = points2DToFPolyline2D(arc);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
			}
            auxRow[ID_FIELD_ID] = ValueFactory.createValue(id);
            auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Circle"));
			String layerName = dwg.getLayerName(entity);
            auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
			int color = entity.getColor();
			int colorByLayer = dwg.getColorByLayer(entity);
			//if (color==0) color ByBlock
			if (color==256) color = colorByLayer;
            auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
        	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(c[2]);
        	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
        	// Attributes
        	/*for (int j=0;j<nAtt;j++) {
				String[] attributes = new String[2];
				attributes = (String[])featureMaker.getAttributes().get(j);
            	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
            	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
            		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
            	}
        	}*/
			addShape(arcc, auxRow);
		} else if (entity instanceof DwgLine) {
			//System.out.println("Encuentra una lnea dentro de un bloque ...");
			double[] p1 = ((DwgLine)entity).getP1();
			double[] p2 = ((DwgLine)entity).getP2();
			Point2D pointAux = new Point2D.Double(p1[0] - bPoint[0], p1[1] - bPoint[1]);
			double laX = insPoint.getX() + ((pointAux.getX()*scale[0])*Math.cos(rot) + (pointAux.getY()*scale[1])*(-1)*Math.sin(rot));
			double laY = insPoint.getY() + ((pointAux.getX()*scale[0])*Math.sin(rot) + (pointAux.getY()*scale[1])*Math.cos(rot));
			Point2D pp1 = new Point2D.Double(laX, laY);
			pointAux = new Point2D.Double(p2[0] - bPoint[0], p2[1] - bPoint[1]);
			laX = insPoint.getX() + ((pointAux.getX()*scale[0])*Math.cos(rot) + (pointAux.getY()*scale[1])*(-1)*Math.sin(rot));
			laY = insPoint.getY() + ((pointAux.getX()*scale[0])*Math.sin(rot) + (pointAux.getY()*scale[1])*Math.cos(rot));
			Point2D pp2 = new Point2D.Double(laX, laY);
			Point2D[] lin = new Point2D[]{pp1, pp2};
			FPolyline2D line;
			if (dwg.isDwg3DFile() && !((DwgLine)entity).isZflag()) {
			    double[][] lin3D = new double[2][3];
		        lin3D[0][0] = lin[0].getX();
		        lin3D[0][1] = lin[0].getY();
		        lin3D[0][2] = p1[2];
		        lin3D[1][0] = lin[1].getX();
		        lin3D[1][1] = lin[1].getY();
		        lin3D[1][2] = p2[2];
			    line = points3DToFPolyline3D(lin3D);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
			} else if (dwg.isDwg3DFile() && ((DwgLine)entity).isZflag()) {
			    double[][] lin3D = new double[2][3];
		        lin3D[0][0] = lin[0].getX();
		        lin3D[0][1] = lin[0].getY();
		        lin3D[0][2] = 0.0;
		        lin3D[1][0] = lin[1].getX();
		        lin3D[1][1] = lin[1].getY();
		        lin3D[1][2] = 0.0;
			    line = points3DToFPolyline3D(lin3D);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
			} else {
			    line = points2DToFPolyline2D(lin);
                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
			}
            auxRow[ID_FIELD_ID] = ValueFactory.createValue(id);
            auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Line"));
			String layerName = dwg.getLayerName(entity);
            auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
			int color = entity.getColor();
			int colorByLayer = dwg.getColorByLayer(entity);
			//if (color==0) color ByBlock
			if (color==256) color = colorByLayer;
            auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
        	if (!((DwgLine)entity).isZflag()) {
                if (p1[2]==p2[2]) auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(p1[2]);
            	else auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
        	} else {
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
        	}
        	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
        	// Attributes
        	/*for (int j=0;j<nAtt;j++) {
				String[] attributes = new String[2];
				attributes = (String[])featureMaker.getAttributes().get(j);
            	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
            	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
            		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
            	}
        	}*/
			addShape(line, auxRow);
		} else if (entity instanceof DwgPolyline2D) {
			//System.out.println("Encuentra una polilnea dentro de un bloque ...");
			Point2D[] points = ((DwgPolyline2D)entity).getPts();
			double elev = ((DwgPolyline2D)entity).getElevation();
			FPolyline2D pline; 
			if (points!=null) {
			    Point2D[] newPoints = new Point2D[points.length];
				for (int i=0;i<points.length;i++) {
				    Point2D pointAux = new Point2D.Double(points[i].getX() - bPoint[0], points[i].getY() - bPoint[1]);
					double laX = insPoint.getX() + ((pointAux.getX()*scale[0])*Math.cos(rot) + (pointAux.getY()*scale[1])*(-1)*Math.sin(rot));
					double laY = insPoint.getY() + ((pointAux.getX()*scale[0])*Math.sin(rot) + (pointAux.getY()*scale[1])*Math.cos(rot));
					newPoints[i] = new Point2D.Double(laX, laY);
				}
				if (dwg.isDwg3DFile()) {
				    double[][] pline3D = new double[newPoints.length][3];
				    for (int j=0;j<newPoints.length;j++) {
				        pline3D[j][0] = newPoints[j].getX();
				        pline3D[j][1] = newPoints[j].getY();
				        pline3D[j][2] = elev;
				    }
				    pline = points3DToFPolyline3D(pline3D);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline3D"));
				} else {
				    pline = points2DToFPolyline2D(newPoints);
	                auxRow[ID_FIELD_FSHAPE] = ValueFactory.createValue(new String("FPolyline2D"));
				}
				auxRow[ID_FIELD_ID] = ValueFactory.createValue(id);
                auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("Polyline2D"));
				String layerName = dwg.getLayerName(entity);
                auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(layerName);
				int color = entity.getColor();
				int colorByLayer = dwg.getColorByLayer(entity);
				//if (color==0) color ByBlock
				if (color==256) color = colorByLayer;
                auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(color);
            	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(elev);
            	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);
            	// Attributes
            	/*for (int j=0;j<nAtt;j++) {
    				String[] attributes = new String[2];
    				attributes = (String[])featureMaker.getAttributes().get(j);
                	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
                	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
                		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
                	}
            	}*/
            	addShape(pline, auxRow);
			}
		} else if (entity instanceof DwgInsert) {
			//System.out.println("Encuentra un insert dentro de un bloque ...");
			double[] p = ((DwgInsert)entity).getInsertionPoint();
			Point2D point = new Point2D.Double(p[0], p[1]);
			double[] newScale = ((DwgInsert)entity).getScale();
			double newRot = ((DwgInsert)entity).getRotation();
			int newBlockHandle = ((DwgInsert)entity).getBlockHeaderHandle();
			Point2D pointAux = new Point2D.Double(point.getX() - bPoint[0], point.getY() - bPoint[1]);
			double laX = insPoint.getX() + ((pointAux.getX()*scale[0])*Math.cos(rot) + (pointAux.getY()*scale[1])*(-1)*Math.sin(rot));
			double laY = insPoint.getY() + ((pointAux.getX()*scale[0])*Math.sin(rot) + (pointAux.getY()*scale[1])*Math.cos(rot));
			Point2D newInsPoint = new Point2D.Double(laX, laY);
			newScale = new double[]{scale[0]*newScale[0], scale[1]*newScale[1], scale[2]*newScale[2]};
			manageInsert(dwgObjects, newInsPoint, newScale, newRot, newBlockHandle, id, auxRow);
			/*auxRow[ID_FIELD_ID] = ValueFactory.createValue(0);
            auxRow[ID_FIELD_ENTITY] = ValueFactory.createValue(new String("dxfEntity"));
            auxRow[ID_FIELD_LAYER] = ValueFactory.createValue(new String("layer"));
            auxRow[ID_FIELD_COLOR] = ValueFactory.createValue(0);
        	auxRow[ID_FIELD_ELEVATION] = ValueFactory.createValue(0.0);
        	auxRow[ID_FIELD_THICKNESS] = ValueFactory.createValue(0.0);*/
        	// Attributes
        	/*for (int j=0;j<nAtt;j++) {
				String[] attributes = new String[2];
				attributes = (String[])featureMaker.getAttributes().get(j);
            	auxRow[9+j] = ValueFactory.createValue(new String((String)attributes[1]));
            	if (!fea.getProp(attributes[0]).equals(attributes[1])) {
            		auxRow[9+j] = ValueFactory.createValue(new String(fea.getProp(attributes[0])));
            	}
        	}*/
        	/*FPoint2D fPoint = createPoint(newInsPoint);
			addShape(fPoint, auxRow);*/
		}
	}
	
	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.drivers.MemoryDriver#accept(java.io.File)
	 */
	public boolean accept(File f) {
		return f.getName().toUpperCase().endsWith("DWG");
	}
	
	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.drivers.MemoryDriver#getShapeType()
	 */
	public int getShapeType() {
		return FShape.MULTI;
	}
	
	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.drivers.MemoryDriver#getName()
	 */
	public String getName() {
		return "gvSIG DWG Memory Driver";
	}
	
	/* (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.drivers.WithDefaultLegend#getDefaultLegend()
	 */
	public Legend getDefaultLegend() {
		return defaultLegend;
	}
	
    /* (non-Javadoc)
     * @see com.iver.cit.gvsig.fmap.drivers.VectorialDriver#getDriverAttributes()
     */
    public DriverAttributes getDriverAttributes() {
        return attr;
    }

    /* (non-Javadoc)
     * @see com.hardcode.gdbms.engine.data.driver.ObjectDriver#getPrimaryKeys()
     */
    public int[] getPrimaryKeys() throws DriverException {
        // TODO Auto-generated method stub
        return null;
    }

    /* (non-Javadoc)
     * @see com.hardcode.gdbms.engine.data.driver.ObjectDriver#write(com.hardcode.gdbms.engine.data.edition.DataWare)
     */
    public void write(DataWare arg0) throws DriverException {
        // TODO Auto-generated method stub
        
    }
}
